library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library ice40up;
use ice40up.components.all;

entity top is
	port (
		clk : in std_logic;
		rxd : out std_logic;
		txd : in std_logic;
		led_blue : out std_logic;
		led_green : out std_logic;
		led_red : out std_logic);
end entity top;

architecture rtl of top is
	constant system_speed : natural := 12e6;
	constant baudrate : natural := 115200;
	constant ram_size : natural := 10 / 2;

	signal rs232_receiver_dat : unsigned(7 downto 0);
	signal rs232_receiver_stb : std_logic;

	signal rs232_sender_dat : unsigned(7 downto 0);
	signal rs232_sender_stb : std_logic;

	signal led_latch_blue : std_logic;
	signal led_latch_green : std_logic;
	signal led_latch_red : std_logic;
	signal counter : natural range 0 to (system_speed / 2);

	signal ram_ad : unsigned(13 downto 0);
	signal ram_di : unsigned(15 downto 0);
	signal ram_maskwe : unsigned(3 downto 0);
	signal ram_we : std_logic;
	signal ram_do : unsigned(15 downto 0);

	signal ebr_data : unsigned(15 downto 0);
	signal ebr_addr : unsigned(11 downto 0);

	signal high_byte : std_logic;
	signal download_done : std_logic;

	signal reset_counter : unsigned(1 downto 0);
	signal reset : std_logic;

	type state_type is (
		wait_for_byte,
		send_data);

	signal state : state_type;

begin

	ram : SP256K
	port map(
		AD => std_logic_vector(ram_ad),
		DI => std_logic_vector(ram_di),
		MASKWE => std_logic_vector(ram_maskwe),
		WE => ram_we,
		CS => '1',
		CK => clk,
		unsigned(DO) => ram_do,
		stdby => '0',
		sleep => '0',
		pwroff_n => '1'
	);

	ebr_inst : ebr
	port map(
		clk_i => clk,
		clk_en_i => '1',
		wr_en_i => '0',
		wr_data_i => (others => '0'),
		addr_i => ebr_addr,
		rd_data_o => ebr_data
	);

	rgb_led : RGB
	generic map(RGB0_CURRENT => "0b000001", RGB1_CURRENT => "0b000001", RGB2_CURRENT => "0b000001")
	port map(
		CURREN => '1',
		RGB0PWM => led_latch_blue,
		RGB1PWM => led_latch_green,
		RGB2PWM => led_latch_red,
		RGBLEDEN => '1',
		RGB0 => led_blue,
		RGB1 => led_green,
		RGB2 => led_red);

	sender : entity rs232_sender
		generic map(system_speed, baudrate)
		port map(
			clk_i => clk,
			dat_i => rs232_sender_dat,
			rst_i => '0',
			stb_i => rs232_sender_stb,
			tx => rxd,
			busy => open);

	receiver : entity rs232_receiver
		generic map(system_speed, baudrate)
		port map(
			clk_i => clk,
			dat_o => rs232_receiver_dat,
			rst_i => reset,
			stb_o => rs232_receiver_stb,
			rx => txd);

	reset_generator : process (clk, reset_counter)
	begin
		if rising_edge(clk) then
			if reset_counter /= 3 then
				reset_counter <= reset_counter + 1;
			end if;
		end if;
		if reset_counter /= 3 then
			reset <= '1';
		else
			reset <= '0';
		end if;
	end process;

	process (clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
				state <= wait_for_byte;
				ram_ad <= (others => '0');
				high_byte <= '0';
				download_done <= '0';
			else
				rs232_sender_stb <= '0';
				ram_we <= '0';
				if download_done = '0' then
					case state is
						-- read chars from RS232 port and write to RAM
						when wait_for_byte =>
							if rs232_receiver_stb = '1' then
								ram_we <= '1';
								if high_byte = '1' then
									ram_di <= rs232_receiver_dat & x"00";
									ram_maskwe <= "1100";
								else
									ram_di <= x"00" & rs232_receiver_dat;
									ram_maskwe <= "0011";
								end if;
								state <= send_data;
							end if;
						when send_data =>
							if high_byte = '1' then
								ram_ad <= ram_ad + 1;
								high_byte <= '0';
								if ram_ad = ram_size - 1 then
									download_done <= '1';
									ram_ad <= (others => '0');
								end if;
							else
								high_byte <= '1';
							end if;
							state <= wait_for_byte;
					end case;
				end if;

				-- blink green LED with 1 Hz and echo RAM contents
				if counter = 0 then
					led_latch_green <= not led_latch_green;
					counter <= system_speed / 2;
					rs232_sender_stb <= '0';
					if download_done = '1' then
						if ram_ad /= ram_size then
							rs232_sender_stb <= '1';
							if high_byte = '1' then
								ram_ad <= ram_ad + 1;
								high_byte <= '0';
								rs232_sender_dat <= ram_do(15 downto 8);
							else
								high_byte <= '1';
								rs232_sender_dat <= ram_do(7 downto 0);
							end if;
						else
							ram_ad <= (others => '0');
							download_done <= '0';
						end if;
					end if;

					led_latch_blue <= ebr_data(0) xor ebr_data(1)xor ebr_data(2)xor ebr_data(3)xor ebr_data(4)xor ebr_data(5)xor ebr_data(6)xor ebr_data(7)xor ebr_data(8)xor ebr_data(9)xor ebr_data(10)xor ebr_data(11)xor ebr_data(12)xor ebr_data(13)xor ebr_data(14)xor ebr_data(15);
					ebr_addr <= ebr_addr + 1;

				else
					counter <= counter - 1;
				end if;
			end if;

			-- red LED on while downloading, green LED blinking while echo, green LED on when done
			led_latch_red <= '0';
			if download_done = '0' then
				if ram_ad = 0 and high_byte = '0' then
					led_latch_green <= '1';
				else
					led_latch_green <= '0';
					led_latch_red <= '1';
				end if;
			end if;
		end if;
	end process;

end architecture rtl;